import unittest

import numpy as np
from six import StringIO

from kafe2.fit._base import ModelFunctionBase, ModelFunctionFormatter
from kafe2.fit.histogram import HistModelFunction
from kafe2.fit.indexed import IndexedModelFunction, IndexedModelFunctionFormatter
from kafe2.fit.io.handle import IOStreamHandle
from kafe2.fit.representation import ModelFunctionYamlReader, ModelFunctionYamlWriter

TEST_MODEL_FUNCTION_HIST = """
type: histogram
python_code: 'x a b -> a * x + b'
"""

TEST_MODEL_FUNCTION_HIST_MISSING_KEYWORD = """
type: histogram
"""

TEST_MODEL_FUNCTION_HIST_EXTRA_KEYWORD = (
    TEST_MODEL_FUNCTION_HIST
    + """
extra_keyword: 3.14
"""
)

TEST_MODEL_FUNCTION_HIST_WITH_FORMATTER = (
    TEST_MODEL_FUNCTION_HIST
    + r"""
model_function_formatter:
    name: linear_model
    latex_name: linear model
    arg_formatters:
      x: r
      a:
        name: alpha
        latex_name: \alpha
      b:
        name: beta
        latex_name: \beta
    expression_string: '{0} * {x} + {1}'
    latex_expression_string: '{0}{x} + {1}'
"""
)


class TestHistModelFunctionYamlRepresenter(unittest.TestCase):
    @staticmethod
    def linear_model(x, a, b):
        return a * x + b

    def setUp(self):
        self._test_x = np.arange(10)
        self._test_a = 2.0
        self._test_b = -1.0
        self._test_y = self.linear_model(self._test_x, self._test_a, self._test_b)

        self._model_function = HistModelFunction(self.linear_model)

        self._roundtrip_stringstream = IOStreamHandle(StringIO())
        self._testfile_stringstream = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_HIST))
        self._testfile_stringstream_with_formatter = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_HIST_WITH_FORMATTER))

        self._roundtrip_streamreader = ModelFunctionYamlReader(self._roundtrip_stringstream)
        self._roundtrip_streamwriter = ModelFunctionYamlWriter(self._model_function, self._roundtrip_stringstream)
        self._testfile_streamreader = ModelFunctionYamlReader(self._testfile_stringstream)
        self._testfile_streamreader_with_formatter = ModelFunctionYamlReader(self._testfile_stringstream_with_formatter)

        self._testfile_stringstream_missing_keyword = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_HIST_MISSING_KEYWORD))
        self._testfile_stringstream_extra_keyword = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_HIST_EXTRA_KEYWORD))
        self._testfile_streamreader_missing_keyword = ModelFunctionYamlReader(self._testfile_stringstream_missing_keyword)
        self._testfile_streamreader_extra_keyword = ModelFunctionYamlReader(self._testfile_stringstream_extra_keyword)

    def test_write_to_roundtrip_stringstream(self):
        self._roundtrip_streamwriter.write()

    def test_read_from_testfile_stream(self):
        _read_model_function = self._testfile_streamreader.read()
        self.assertTrue(isinstance(_read_model_function, HistModelFunction))
        self.assertTrue(np.allclose(_read_model_function.func(self._test_x, self._test_a, self._test_b), self._test_y))

    def test_read_from_testfile_stream_missing_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_missing_keyword.read()

    def test_read_from_testfile_stream_extra_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_extra_keyword.read()

    def test_read_from_testfile_stream_with_formatter(self):
        _read_model_function = self._testfile_streamreader_with_formatter.read()
        self.assertTrue(isinstance(_read_model_function, HistModelFunction))
        self.assertTrue(np.allclose(_read_model_function.func(self._test_x, self._test_a, self._test_b), self._test_y))
        _read_formatter = _read_model_function.formatter
        self.assertTrue(isinstance(_read_formatter, ModelFunctionFormatter))
        _read_arg_formatters = _read_formatter.arg_formatters
        self.assertTrue(_read_formatter.name == "linear_model")
        self.assertTrue(_read_formatter.latex_name == "linear model")
        self.assertTrue(_read_arg_formatters[0].name == "x")
        self.assertTrue(_read_arg_formatters[0].latex_name == r"{r}")
        self.assertTrue(_read_arg_formatters[1].name == "alpha")
        self.assertTrue(_read_arg_formatters[1].latex_name == r"{\alpha}")
        self.assertTrue(_read_arg_formatters[2].name == "beta")
        self.assertTrue(_read_arg_formatters[2].latex_name == r"{\beta}")
        self.assertTrue(_read_formatter.expression_format_string == "{0} * {x} + {1}")
        self.assertTrue(_read_formatter.latex_expression_format_string == "{0}{x} + {1}")

    def test_round_trip_with_stringstream(self):
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_model_function = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_model_function, HistModelFunction))

        self.assertTrue(np.allclose(self._test_y, _read_model_function.func(self._test_x, self._test_a, self._test_b)))

        _given_formatter = self._model_function.formatter
        _read_formatter = _read_model_function.formatter

        self.assertTrue(isinstance(_read_formatter, ModelFunctionFormatter))

        _given_arg_formatters = _given_formatter.arg_formatters
        _read_arg_formatters = _read_formatter.arg_formatters

        self.assertTrue(_read_formatter.name == _given_formatter.name)
        self.assertTrue(_read_formatter.latex_name == _given_formatter.latex_name)
        self.assertTrue(_read_arg_formatters[0].name == _given_arg_formatters[0].name)
        self.assertTrue(_read_arg_formatters[0].latex_name == _given_arg_formatters[0].latex_name)
        self.assertTrue(_read_arg_formatters[1].name == _given_arg_formatters[1].name)
        self.assertTrue(_read_arg_formatters[1].latex_name == _given_arg_formatters[1].latex_name)
        self.assertTrue(_read_arg_formatters[2].name == _given_arg_formatters[2].name)
        self.assertTrue(_read_arg_formatters[2].latex_name == _given_arg_formatters[2].latex_name)
        self.assertTrue(_read_formatter.expression_format_string == _given_formatter.expression_format_string)
        self.assertTrue(_read_formatter.latex_expression_format_string == _given_formatter.latex_expression_format_string)


TEST_MODEL_FUNCTION_INDEXED = """
type: indexed
python_code: |
    def linear_model(a, b):
        return a * np.arange(10) + b
"""

TEST_MODEL_FUNCTION_INDEXED_MISSING_KEYWORD = """
type: indexed
"""

TEST_MODEL_FUNCTION_INDEXED_EXTRA_KEYWORD = (
    TEST_MODEL_FUNCTION_INDEXED
    + """
extra_keyword: 3.14
"""
)

TEST_MODEL_FUNCTION_INDEXED_WITH_FORMATTER = (
    TEST_MODEL_FUNCTION_INDEXED
    + r"""
model_function_formatter:
    name: linear_model
    latex_name: linear model
    index_name: r
    latex_index_name: r
    arg_formatters:
      a:
        name: alpha
        latex_name: \alpha
      b:
        name: beta
        latex_name: \beta
    expression_string: '{0} * {r} + {1}'
    latex_expression_string: '{0}{r} + {1}'
"""
)


class TestIndexedModelFunctionYamlRepresenter(unittest.TestCase):
    TEST_X = np.arange(10)

    @staticmethod
    def linear_model(a, b):
        # TODO handle
        # return a * TestIndexedModelFunctionYamlRepresenter.TEST_X + b
        return a * np.arange(10) + b

    def setUp(self):
        self._test_a = 2.0
        self._test_b = -1.0
        self._test_y = self.linear_model(self._test_a, self._test_b)

        self._model_function = IndexedModelFunction(self.linear_model)

        self._roundtrip_stringstream = IOStreamHandle(StringIO())
        self._testfile_stringstream = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_INDEXED))
        self._testfile_stringstream_with_formatter = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_INDEXED_WITH_FORMATTER))

        self._roundtrip_streamreader = ModelFunctionYamlReader(self._roundtrip_stringstream)
        self._roundtrip_streamwriter = ModelFunctionYamlWriter(self._model_function, self._roundtrip_stringstream)
        self._testfile_streamreader = ModelFunctionYamlReader(self._testfile_stringstream)
        self._testfile_streamreader_with_formatter = ModelFunctionYamlReader(self._testfile_stringstream_with_formatter)

        self._testfile_stringstream_missing_keyword = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_INDEXED_MISSING_KEYWORD))
        self._testfile_stringstream_extra_keyword = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_INDEXED_EXTRA_KEYWORD))
        self._testfile_streamreader_missing_keyword = ModelFunctionYamlReader(self._testfile_stringstream_missing_keyword)
        self._testfile_streamreader_extra_keyword = ModelFunctionYamlReader(self._testfile_stringstream_extra_keyword)

    def test_write_to_roundtrip_stringstream(self):
        self._roundtrip_streamwriter.write()

    def test_read_from_testfile_stream(self):
        _read_model_function = self._testfile_streamreader.read()
        self.assertTrue(isinstance(_read_model_function, IndexedModelFunction))
        self.assertTrue(np.allclose(_read_model_function.func(self._test_a, self._test_b), self._test_y))

    def test_read_from_testfile_stream_missing_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_missing_keyword.read()

    def test_read_from_testfile_stream_extra_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_extra_keyword.read()

    def test_read_from_testfile_stream_with_formatter(self):
        _read_model_function = self._testfile_streamreader_with_formatter.read()
        self.assertTrue(isinstance(_read_model_function, IndexedModelFunction))
        self.assertTrue(np.allclose(_read_model_function.func(self._test_a, self._test_b), self._test_y))
        _read_formatter = _read_model_function.formatter
        self.assertTrue(isinstance(_read_formatter, IndexedModelFunctionFormatter))
        _read_arg_formatters = _read_formatter.arg_formatters
        self.assertTrue(_read_formatter.name == "linear_model")
        self.assertTrue(_read_formatter.latex_name == "linear model")
        self.assertTrue(_read_formatter.index_name == "r")
        self.assertTrue(_read_formatter.latex_index_name == r"{r}")
        self.assertTrue(_read_arg_formatters[0].name == "alpha")
        self.assertTrue(_read_arg_formatters[0].latex_name == r"{\alpha}")
        self.assertTrue(_read_arg_formatters[1].name == "beta")
        self.assertTrue(_read_arg_formatters[1].latex_name == r"{\beta}")
        self.assertTrue(_read_formatter.expression_format_string == "{0} * {r} + {1}")
        self.assertTrue(_read_formatter.latex_expression_format_string == "{0}{r} + {1}")

    def test_round_trip_with_stringstream(self):
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_model_function = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_model_function, IndexedModelFunction))

        self.assertTrue(np.allclose(self._test_y, _read_model_function.func(self._test_a, self._test_b)))

        _given_formatter = self._model_function.formatter
        _read_formatter = _read_model_function.formatter

        self.assertTrue(isinstance(_read_formatter, IndexedModelFunctionFormatter))

        _given_arg_formatters = _given_formatter.arg_formatters
        _read_arg_formatters = _read_formatter.arg_formatters

        self.assertTrue(_read_formatter.name == _given_formatter.name)
        self.assertTrue(_read_formatter.latex_name == _given_formatter.latex_name)
        self.assertTrue(_read_formatter.index_name == _given_formatter.index_name)
        self.assertTrue(_read_formatter.latex_index_name == _given_formatter.latex_index_name)
        self.assertTrue(_read_arg_formatters[0].name == _given_arg_formatters[0].name)
        self.assertTrue(_read_arg_formatters[0].latex_name == _given_arg_formatters[0].latex_name)
        self.assertTrue(_read_arg_formatters[1].name == _given_arg_formatters[1].name)
        self.assertTrue(_read_arg_formatters[1].latex_name == _given_arg_formatters[1].latex_name)
        self.assertTrue(_read_formatter.expression_format_string == _given_formatter.expression_format_string)
        self.assertTrue(_read_formatter.latex_expression_format_string == _given_formatter.latex_expression_format_string)


TEST_MODEL_FUNCTION_BASE = """
python_code: 'x a b -> a * x + b'
"""

TEST_MODEL_FUNCTION_BASE_MISSING_KEYWORD = """
type: base
"""

TEST_MODEL_FUNCTION_XY_EXTRA_KEYWORD = (
    TEST_MODEL_FUNCTION_BASE
    + """
extra_keyword: 3.14
"""
)

TEST_MODEL_FUNCTION_XY_WITH_FORMATTER = (
    TEST_MODEL_FUNCTION_BASE
    + r"""
model_function_formatter:
    name: linear_model
    latex_name: linear model
    arg_formatters:
      x: r
      a:
        name: alpha
        latex_name: \alpha
      b:
        name: beta
        latex_name: \beta
    expression_string: '{0} * {x} + {1}'
    latex_expression_string: '{0}{x} + {1}'
"""
)


class TestModelFunctionBaseYamlRepresenter(unittest.TestCase):
    @staticmethod
    def linear_model(x, a, b):
        return a * x + b

    def setUp(self):
        self._test_x = np.arange(10)
        self._test_a = 2.0
        self._test_b = -1.0
        self._test_y = self.linear_model(self._test_x, self._test_a, self._test_b)

        self._model_function = ModelFunctionBase(self.linear_model)

        self._roundtrip_stringstream = IOStreamHandle(StringIO())
        self._testfile_stringstream = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_BASE))
        self._testfile_stringstream_with_formatter = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_XY_WITH_FORMATTER))

        self._roundtrip_streamreader = ModelFunctionYamlReader(self._roundtrip_stringstream)
        self._roundtrip_streamwriter = ModelFunctionYamlWriter(self._model_function, self._roundtrip_stringstream)
        self._testfile_streamreader = ModelFunctionYamlReader(self._testfile_stringstream)
        self._testfile_streamreader_with_formatter = ModelFunctionYamlReader(self._testfile_stringstream_with_formatter)

        self._testfile_stringstream_missing_keyword = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_BASE_MISSING_KEYWORD))
        self._testfile_stringstream_extra_keyword = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_XY_EXTRA_KEYWORD))
        self._testfile_streamreader_missing_keyword = ModelFunctionYamlReader(self._testfile_stringstream_missing_keyword)
        self._testfile_streamreader_extra_keyword = ModelFunctionYamlReader(self._testfile_stringstream_extra_keyword)

    def test_write_to_roundtrip_stringstream(self):
        self._roundtrip_streamwriter.write()

    def test_read_from_testfile_stream(self):
        _read_model_function = self._testfile_streamreader.read()
        self.assertTrue(isinstance(_read_model_function, ModelFunctionBase))
        self.assertTrue(np.allclose(_read_model_function.func(self._test_x, self._test_a, self._test_b), self._test_y))

    def test_read_from_testfile_stream_missing_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_missing_keyword.read()

    def test_read_from_testfile_stream_extra_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_extra_keyword.read()

    def test_read_from_testfile_stream_with_formatter(self):
        _read_model_function = self._testfile_streamreader_with_formatter.read()
        self.assertTrue(isinstance(_read_model_function, ModelFunctionBase))
        self.assertTrue(np.allclose(_read_model_function.func(self._test_x, self._test_a, self._test_b), self._test_y))
        _read_formatter = _read_model_function.formatter
        self.assertTrue(isinstance(_read_formatter, ModelFunctionFormatter))
        _read_arg_formatters = _read_formatter.arg_formatters
        self.assertTrue(_read_formatter.name == "linear_model")
        self.assertTrue(_read_formatter.latex_name == "linear model")
        self.assertTrue(_read_arg_formatters[0].name == "x")
        self.assertTrue(_read_arg_formatters[0].latex_name == r"{r}")
        self.assertTrue(_read_arg_formatters[1].name == "alpha")
        self.assertTrue(_read_arg_formatters[1].latex_name == r"{\alpha}")
        self.assertTrue(_read_arg_formatters[2].name == "beta")
        self.assertTrue(_read_arg_formatters[2].latex_name == r"{\beta}")
        self.assertTrue(_read_formatter.expression_format_string == "{0} * {x} + {1}")
        self.assertTrue(_read_formatter.latex_expression_format_string == "{0}{x} + {1}")

    def test_round_trip_with_stringstream(self):
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_model_function = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_model_function, ModelFunctionBase))

        self.assertTrue(np.allclose(self._test_y, _read_model_function.func(self._test_x, self._test_a, self._test_b)))

        _given_formatter = self._model_function.formatter
        _read_formatter = _read_model_function.formatter

        self.assertTrue(isinstance(_read_formatter, ModelFunctionFormatter))

        _given_arg_formatters = _given_formatter.arg_formatters
        _read_arg_formatters = _read_formatter.arg_formatters

        self.assertTrue(_read_formatter.name == _given_formatter.name)
        self.assertTrue(_read_formatter.latex_name == _given_formatter.latex_name)
        self.assertTrue(_read_arg_formatters[0].name == _given_arg_formatters[0].name)
        self.assertTrue(_read_arg_formatters[0].latex_name == _given_arg_formatters[0].latex_name)
        self.assertTrue(_read_arg_formatters[1].name == _given_arg_formatters[1].name)
        self.assertTrue(_read_arg_formatters[1].latex_name == _given_arg_formatters[1].latex_name)
        self.assertTrue(_read_formatter.expression_format_string == _given_formatter.expression_format_string)
        self.assertTrue(_read_formatter.latex_expression_format_string == _given_formatter.latex_expression_format_string)



import abc

import six

# import formatter classes
from ..._base.format import ModelFunctionFormatter
from ...indexed import IndexedModelFunctionFormatter
from .._base import GenericDReprBase

__all__ = ["ModelFunctionFormatterDReprBase", "ParameterFormatterDReprBase"]


@six.add_metaclass(abc.ABCMeta)
class ModelFunctionFormatterDReprBase(GenericDReprBase):
    BASE_OBJECT_TYPE_NAME = "model_function_formatter"

    _CLASS_TO_OBJECT_TYPE_NAME = {
        ModelFunctionFormatter: "base",
        IndexedModelFunctionFormatter: "indexed",
    }
    _OBJECT_TYPE_NAME_TO_CLASS = {
        "base": ModelFunctionFormatter,
        "histogram": ModelFunctionFormatter,
        "indexed": IndexedModelFunctionFormatter,
        "unbinned": ModelFunctionFormatter,
        "xy": ModelFunctionFormatter,
    }

    def __init__(self, model_function_formatter=None):
        self._kafe_object = model_function_formatter
        super(ModelFunctionFormatterDReprBase, self).__init__()


@six.add_metaclass(abc.ABCMeta)
class ParameterFormatterDReprBase(GenericDReprBase):
    BASE_OBJECT_TYPE_NAME = "parameter_formatter"

    def __init__(self, model_parameter_formatter=None):
        self._kafe_object = model_parameter_formatter
        super(ParameterFormatterDReprBase, self).__init__()



from .._base import (
    CostFunction_Chi2,
    CostFunction_GaussApproximation,
    CostFunction_NegLogLikelihood,
)

__all__ = [
    "XYCostFunction_Chi2",
    "XYCostFunction_NegLogLikelihood",
    "XYCostFunction_GaussApproximation",
]


class XYCostFunction_Chi2(CostFunction_Chi2):
    def __init__(
        self,
        errors_to_use="covariance",
        fallback_on_singular=True,
        axes_to_use="xy",
        add_constraint_cost=True,
        add_determinant_cost=True,
    ):
        """Built-in least-squares cost function for *xy* data.

        :param errors_to_use: Which errors to use when calculating :math:`\\chi^2`. This is either
            `'covariance'``, ``'pointwise'`` or :py:obj:`None`.
        :type errors_to_use: str or None
        :param axes_to_use: The errors for the given axes are taken into account when calculating
            :math:`\\chi^2`. Either ``'y'`` or ``'xy'``
        :param bool add_constraint_cost: If :py:obj:`True`, automatically add the cost for kafe2
            constraints.
        :param bool add_determinant_cost: If :py:obj:`True`, automatically increase the cost
            function value by the logarithm of the determinant of the covariance matrix to reduce
            bias.
        """
        self._DATA_NAME = "y_data"
        self._MODEL_NAME = "y_model"
        if axes_to_use.lower() == "y":
            self._COV_MAT_CHOLESKY_NAME = "y_total_cov_mat_cholesky"
            self._ERROR_NAME = "y_total_error"
        elif axes_to_use.lower() == "xy":
            self._COV_MAT_CHOLESKY_NAME = "total_cov_mat_cholesky"
            self._ERROR_NAME = "total_error"
        else:
            raise ValueError("Unknown value '%s' for 'axes_to_use': must be one of ('xy', 'y')")
        super(XYCostFunction_Chi2, self).__init__(
            errors_to_use=errors_to_use,
            fallback_on_singular=fallback_on_singular,
            add_constraint_cost=add_constraint_cost,
            add_determinant_cost=add_determinant_cost,
        )


class XYCostFunction_NegLogLikelihood(CostFunction_NegLogLikelihood):
    def __init__(self, data_point_distribution="poisson", ratio=False, axes_to_use="xy"):
        self._DATA_NAME = "y_data"
        self._MODEL_NAME = "y_model"
        if axes_to_use.lower() == "y":
            self._ERROR_NAME = "y_total_error"
        elif axes_to_use.lower() == "xy":
            self._ERROR_NAME = "total_error"
        else:
            raise ValueError("Unknown value '%s' for 'axes_to_use': must be one of ('xy', 'y')")
        super(XYCostFunction_NegLogLikelihood, self).__init__(data_point_distribution=data_point_distribution, ratio=ratio)


class XYCostFunction_GaussApproximation(CostFunction_GaussApproximation):
    def __init__(
        self,
        errors_to_use="covariance",
        axes_to_use="xy",
        add_constraint_cost=True,
        add_determinant_cost=True,
    ):
        """
        Built-in Gaussian approximation of the Poisson negative log-likelihood cost function for
        *xy* data.

        :param errors_to_use: Which errors to use when calculating :math:`\\chi^2`. This is either
            `'covariance'``, ``'pointwise'``.
        :type errors_to_use: str
        :param axes_to_use: The errors for the given axes are taken into account when calculating
            :math:`\\chi^2`. Either ``'y'`` or ``'xy'``
        :param bool add_constraint_cost: If :py:obj:`True`, automatically add the cost for kafe2
            constraints.
        :param bool add_determinant_cost: If :py:obj:`True`, automatically increase the cost
            function value by the logarithm of the determinant of the covariance matrix to reduce
            bias.
        """
        self._DATA_NAME = "y_data"
        self._MODEL_NAME = "y_model"
        if axes_to_use.lower() == "y":
            self._COV_MAT_CHOLESKY_NAME = "y_total_cov_mat_cholesky"
            self._ERROR_NAME = "y_total_error"
        elif axes_to_use.lower() == "xy":
            self._COV_MAT_CHOLESKY_NAME = "total_cov_mat_cholesky"
            self._ERROR_NAME = "total_error"
        else:
            raise ValueError("Unknown value '%s' for 'axes_to_use': must be one of ('xy', 'y')")
        super().__init__(
            errors_to_use=errors_to_use,
            add_constraint_cost=add_constraint_cost,
            add_determinant_cost=add_determinant_cost,
        )


STRING_TO_COST_FUNCTION = {
    "chi2": (XYCostFunction_Chi2, {}),
    "chi_2": (XYCostFunction_Chi2, {}),
    "chisquared": (XYCostFunction_Chi2, {}),
    "chi_squared": (XYCostFunction_Chi2, {}),
    "chi2_no_errors": (XYCostFunction_Chi2, {"errors_to_use": None, "add_determinant_cost": False}),
    "chi2_pointwise": (XYCostFunction_Chi2, {"errors_to_use": "pointwise"}),
    "chi2_pointwise_errors": (XYCostFunction_Chi2, {"errors_to_use": "pointwise"}),
    "chi2_covariance": (XYCostFunction_Chi2, {"errors_to_use": "covariance"}),
    "nll": (XYCostFunction_NegLogLikelihood, {"ratio": False}),
    "poisson": (
        XYCostFunction_NegLogLikelihood,
        {"data_point_distribution": "poisson", "ratio": False},
    ),
    "nll-poisson": (
        XYCostFunction_NegLogLikelihood,
        {"data_point_distribution": "poisson", "ratio": False},
    ),
    "nll_poisson": (
        XYCostFunction_NegLogLikelihood,
        {"data_point_distribution": "poisson", "ratio": False},
    ),
    "nllpoisson": (
        XYCostFunction_NegLogLikelihood,
        {"data_point_distribution": "poisson", "ratio": False},
    ),
    "nll-gaussian": (
        XYCostFunction_NegLogLikelihood,
        {"data_point_distribution": "gaussian", "ratio": False},
    ),
    "nll_gaussian": (
        XYCostFunction_NegLogLikelihood,
        {"data_point_distribution": "gaussian", "ratio": False},
    ),
    "nllgaussiann": (
        XYCostFunction_NegLogLikelihood,
        {"data_point_distribution": "gaussian", "ratio": False},
    ),
    "negloglikelihood": (XYCostFunction_NegLogLikelihood, {"ratio": False}),
    "neg_log_likelihood": (XYCostFunction_NegLogLikelihood, {"ratio": False}),
    "nllr": (CostFunction_NegLogLikelihood, {"ratio": True}),
    "nllr-poisson": (
        XYCostFunction_NegLogLikelihood,
        {"data_point_distribution": "poisson", "ratio": True},
    ),
    "nllr_poisson": (
        XYCostFunction_NegLogLikelihood,
        {"data_point_distribution": "poisson", "ratio": True},
    ),
    "nllrpoisson": (
        XYCostFunction_NegLogLikelihood,
        {"data_point_distribution": "poisson", "ratio": True},
    ),
    "nllr-gaussian": (
        XYCostFunction_NegLogLikelihood,
        {"data_point_distribution": "gaussian", "ratio": True},
    ),
    "nllr_gaussian": (
        XYCostFunction_NegLogLikelihood,
        {"data_point_distribution": "gaussian", "ratio": True},
    ),
    "nllrgaussian": (
        XYCostFunction_NegLogLikelihood,
        {"data_point_distribution": "gaussian", "ratio": True},
    ),
    "negloglikelihoodratio": (XYCostFunction_NegLogLikelihood, {"ratio": True}),
    "neg_log_likelihood_ratio": (XYCostFunction_NegLogLikelihood, {"ratio": True}),
    "gauss-approximation": (XYCostFunction_GaussApproximation, {}),
    "gauss_approximation": (XYCostFunction_GaussApproximation, {}),
    "gauss_approximation_covariance": (
        XYCostFunction_GaussApproximation,
        {"errors_to_use": "covariance"},
    ),
    "gauss_approximation_pointwise": (
        XYCostFunction_GaussApproximation,
        {"errors_to_use": "pointwise"},
    ),
    "gauss_approximation_pointwise_errors": (
        XYCostFunction_GaussApproximation,
        {"errors_to_use": "pointwise"},
    ),
}



"""This submodule contains objects for handling different representations of models and model functions.

:synopsis: This submodule contains objects for handling different representations of models and model functions.

.. moduleauthor:: Johannes Gäßler <johannes.gaessler@cern.ch>
"""

# flake8: noqa F401, F403 (imported but unused, used but unable to detect undefined names)

from ._base import *
from .yaml_drepr import *



#!/usr/bin/env python3

import os
import shutil
import unittest

import numpy as np

from kafe2 import XYFit, function_library, xy_fit


class TestWrapperCallableXY(unittest.TestCase):
    def _assert_results_equal(self, results_1, results_2):
        self.assertEqual(results_1["did_fit"], results_2["did_fit"])
        self.assertAlmostEqual(results_1["cost"], results_2["cost"])
        self.assertAlmostEqual(results_1["goodness_of_fit"], results_2["goodness_of_fit"])
        self.assertAlmostEqual(results_1["chi2_probability"], results_2["chi2_probability"])
        self.assertAlmostEqual(results_1["gof/ndf"], results_2["gof/ndf"])
        self.assertEqual(results_1["ndf"], results_2["ndf"])
        self.assertTrue(
            np.allclose(
                list(results_1["parameter_values"].values()),
                list(results_2["parameter_values"].values()),
            )
        )
        self.assertTrue(
            np.allclose(
                list(results_1["parameter_errors"].values()),
                list(results_2["parameter_errors"].values()),
            )
        )
        self.assertTrue(np.allclose(results_1["parameter_cor_mat"], results_2["parameter_cor_mat"], rtol=1e-4))
        self.assertTrue(np.allclose(results_1["parameter_cov_mat"], results_2["parameter_cov_mat"], rtol=1e-4))
        try:
            self.assertTrue(
                np.allclose(
                    list(results_1["asymmetric_parameter_errors"].values()),
                    list(results_2["asymmetric_parameter_errors"].values()),
                )
            )
        except AttributeError:
            self.assertEqual(results_1["asymmetric_parameter_errors"], None)
            self.assertEqual(results_2["asymmetric_parameter_errors"], None)

    def _assert_wrapper_equal(
        self,
        model_function=function_library.linear_model,
        p0=None,
        dp0=None,
        rel_to_model=True,
        limits=None,
        constraints=None,
        profile=None,
        save=True,
        test_matrix=False,
    ):
        _fit = XYFit([self._x_data, self._y_data], model_function)
        _error_ref = "model" if rel_to_model else "data"
        if p0 is not None:
            _fit.set_all_parameter_values(p0)
        if dp0 is not None:
            _fit.parameter_errors = dp0
        _fit.add_error("x", self._x_error)
        _fit.add_error("y", self._y_error)
        _fit.add_error("x", self._x_error_rel, relative=True)
        _fit.add_error("y", self._y_error_rel, relative=True, reference=_error_ref)
        _fit.add_error("x", self._x_error_cor, correlation=1.0)
        _fit.add_error("y", self._y_error_cor[0], correlation=1.0)
        _fit.add_error("y", self._y_error_cor[1], correlation=1.0)
        _fit.add_error("x", self._x_error_cor_rel[0], correlation=1.0, relative=True)
        _fit.add_error("x", self._x_error_cor_rel[1], correlation=1.0, relative=True)
        _fit.add_error("y", self._y_error_cor_rel, correlation=1.0, relative=True, reference=_error_ref)
        if limits is not None:
            _limits_manual = [limits] if isinstance(limits[0], str) else limits
            for _limit in _limits_manual:
                _fit.limit_parameter(*_limit)
        if constraints is not None:
            _constraints_manual = [constraints] if isinstance(constraints[0], str) else constraints
            for _constraint in _constraints_manual:
                _fit.add_parameter_constraint(*_constraint)
        _profile_manual = profile is None or profile
        _result_1 = _fit.do_fit(asymmetric_parameter_errors=_profile_manual)

        if test_matrix:
            self._x_error = np.eye(10) * self._x_error**2
            self._x_error += np.sum(np.square(self._x_error_cor))
            self._x_error_cor = None
            self._y_error = np.eye(10) * self._y_error**2
            self._y_error += np.sum(np.square(self._y_error_cor))
            self._y_error_cor = None

        _result_2 = xy_fit(
            model_function,
            self._x_data,
            self._y_data,
            p0,
            dp0,
            x_error=self._x_error,
            y_error=self._y_error,
            x_error_rel=self._x_error_rel,
            y_error_rel=self._y_error_rel,
            x_error_cor=self._x_error_cor,
            y_error_cor=self._y_error_cor,
            x_error_cor_rel=self._x_error_cor_rel,
            y_error_cor_rel=self._y_error_cor_rel,
            errors_rel_to_model=rel_to_model,
            limits=limits,
            constraints=constraints,
            profile=profile,
            save=save,
        )
        self._assert_results_equal(_result_1, _result_2)

    def setUp(self):
        np.random.seed(12345)
        self._true_pars = [1.2, 3.4]

        self._x_error = 0.1
        self._y_error = np.array([0.1] * 5 + [0.2] * 5)
        self._x_error_rel = np.array([0.02] * 5 + [0.0] * 5)
        self._y_error_rel = 0.02
        self._x_error_cor = 0.05
        self._y_error_cor = [0.05, 0.1]
        self._x_error_cor_rel = [0.01, 0.02]
        self._y_error_cor_rel = 0.01

        self._x_data = np.arange(10)
        self._y_data = function_library.linear_model(self._x_data + 0.2 * np.random.randn(10), *self._true_pars) + 0.5 * np.random.randn(10)

    def tearDown(self):
        if os.path.exists("results"):
            shutil.rmtree("results")

    def test_default_model_function(self):
        self._assert_wrapper_equal()

    def test_quadratic_model_function(self):
        self._assert_wrapper_equal(function_library.quadratic_model)

    def test_set_p0(self):
        self._assert_wrapper_equal(function_library.exponential_model, p0=[2.0, 2.0], dp0=[0.01, 0.01])

    def test_rel_to_data(self):
        self._assert_wrapper_equal(rel_to_model=False)

    def test_limits(self):
        self._assert_wrapper_equal(limits=["a", -2, 2])
        self._assert_wrapper_equal(limits=[["a", -2, 2], ["b", -2, 2]])

    def test_constraints(self):
        self._assert_wrapper_equal(constraints=["a", -2, 2])
        self._assert_wrapper_equal(constraints=[["a", -2, 2], ["b", -2, 2]])

    def test_explicit_profile(self):
        self._assert_wrapper_equal(profile=False)
        self._assert_wrapper_equal(profile=True)

    def test_matrix(self):
        self._assert_wrapper_equal(test_matrix=True)

    def test_save(self):
        self._assert_wrapper_equal()
        self.assertEqual(len(os.listdir("results")), 2)
        self._assert_wrapper_equal(save=False)
        self.assertEqual(len(os.listdir("results")), 2)
        self._assert_wrapper_equal()
        self.assertEqual(len(os.listdir("results")), 4)



import abc

import six

from ....core.constraint import (
    GaussianMatrixParameterConstraint,
    GaussianSimpleParameterConstraint,
)
from .._base import GenericDReprBase

__all__ = ["ConstraintDReprBase"]


@six.add_metaclass(abc.ABCMeta)
class ConstraintDReprBase(GenericDReprBase):
    BASE_OBJECT_TYPE_NAME = "constraint"

    _CLASS_TO_OBJECT_TYPE_NAME = {
        GaussianSimpleParameterConstraint: "simple",
        GaussianMatrixParameterConstraint: "matrix",
    }
    _OBJECT_TYPE_NAME_TO_CLASS = {
        "simple": GaussianSimpleParameterConstraint,
        "matrix": GaussianMatrixParameterConstraint,
    }

    def __init__(self, constraint=None):
        self._kafe_object = constraint
        super(ConstraintDReprBase, self).__init__()



import unittest

from kafe2.test.core.minimizers._base import AbstractMinimizerTest

_cannot_import_IMinuit = False
try:
    from kafe2.core.minimizers.iminuit_minimizer import MinimizerIMinuit
except (ImportError, SyntaxError):
    _cannot_import_IMinuit = True


@unittest.skipIf(_cannot_import_IMinuit, "Cannot import iminuit")
class TestMinimizerIMinuit(AbstractMinimizerTest, unittest.TestCase):
    def _get_minimizer(self, parameter_names, parameter_values, parameter_errors, function_to_minimize):
        return MinimizerIMinuit(
            parameter_names=parameter_names,
            parameter_values=parameter_values,
            parameter_errors=parameter_errors,
            function_to_minimize=function_to_minimize,
        )

    @property
    def _expected_tolerance(self):
        return 1e-2



import unittest

import numpy as np
from scipy.special import factorial

from kafe2.core.constraint import (
    GaussianMatrixParameterConstraint,
    GaussianSimpleParameterConstraint,
)
from kafe2.fit._base.cost import *
from kafe2.fit.histogram.cost import *
from kafe2.fit.indexed.cost import *
from kafe2.fit.xy.cost import *

# flake8: noqa F405
# function may be undefined or defined from ...
# this applies to nearly each cost function in this document.


class TestCostBuiltin(unittest.TestCase):
    CHI2_COST_FUNCTION = CostFunction_Chi2
    NLL_COST_FUNCTION = CostFunction_NegLogLikelihood
    GAUSS_APPROXIMATION_COST_FUNCTION = CostFunction_GaussApproximation

    def setUp(self):
        self._data_chi2 = np.array([-0.5, 2.1, 8.9])
        self._model_chi2 = np.array([5.7, 8.4, -2.3])
        self._data_poisson = np.array([0.0, 2.0, 9.0])
        self._model_poisson = np.array([5.7, 8.4, 2.3])
        self._res = self._data_chi2 - self._model_chi2
        self._cov_mat = np.array([[1.0, 0.1, 0.2], [0.1, 2.0, 0.3], [0.2, 0.3, 3.0]])
        self._cov_mat_cholesky = np.linalg.cholesky(self._cov_mat)
        self._cov_mat_inv = np.linalg.inv(self._cov_mat)
        self._pointwise_errors = np.sqrt(np.diag(self._cov_mat))
        self._data_poisson_large = np.array([100000, 200000, 300000])
        self._model_poisson_large = np.array([99990, 200010, 300000])

        self._cov_mat_log_det = np.log(np.linalg.det(self._cov_mat))
        self._total_error_squared_log_sum = 2.0 * np.sum(np.log(self._pointwise_errors))
        self._cost_chi2_cov_mat = self._res.dot(self._cov_mat_inv).dot(self._res) + self._cov_mat_log_det
        self._cost_chi2_pointwise = np.sum((self._res / self._pointwise_errors) ** 2) + self._total_error_squared_log_sum
        self._cost_chi2_no_errors = np.sum(self._res**2)
        self._cost_nll_gaussian = (
            self._cost_chi2_pointwise + 2.0 * np.sum(np.log(np.sqrt((2.0 * np.pi)) * self._pointwise_errors)) - self._total_error_squared_log_sum
        )
        self._cost_nll_poisson = -2.0 * np.sum(
            np.log(self._model_poisson**self._data_poisson) - np.log(factorial(self._data_poisson)) - self._model_poisson
        )
        self._cost_nllr_poisson = self._cost_nll_poisson + 2.0 * np.sum(
            np.log(self._data_poisson**self._data_poisson) - np.log(factorial(self._data_poisson)) - self._data_poisson
        )
        self._par_vals = np.array([12.3, 0.001, -1.9])
        self._simple_constraint = GaussianSimpleParameterConstraint(index=0, value=10.0, uncertainty=2.5)
        self._matrix_constraint = GaussianMatrixParameterConstraint(
            indices=(0, 1, 2),
            values=(1.0, 2.0, 3.0),
            matrix=[[1.5, 0.1, 0.1], [0.1, 2.2, 0.1], [0.1, 0.1, 0.3]],
        )
        self._par_constraints = [self._simple_constraint, self._matrix_constraint]
        self._par_cost = self._simple_constraint.cost(self._par_vals) + self._matrix_constraint.cost(self._par_vals)

    def test_chi2_no_errors(self):
        self.assertAlmostEqual(
            self._cost_chi2_no_errors,
            self.CHI2_COST_FUNCTION(errors_to_use=None, add_determinant_cost=False)(self._data_chi2, self._model_chi2, None, None),
        )
        self.assertAlmostEqual(
            self._cost_chi2_no_errors + self._par_cost,
            self.CHI2_COST_FUNCTION(errors_to_use=None, add_determinant_cost=False)(
                self._data_chi2, self._model_chi2, self._par_vals, self._par_constraints
            ),
        )

    def test_chi2_pointwise(self):
        self.assertAlmostEqual(
            self._cost_chi2_pointwise,
            self.CHI2_COST_FUNCTION(errors_to_use="pointwise")(
                self._data_chi2,
                self._model_chi2,
                self._pointwise_errors,
                None,
                None,
                self._total_error_squared_log_sum,
            ),
        )
        self.assertAlmostEqual(
            self._cost_chi2_pointwise + self._par_cost,
            self.CHI2_COST_FUNCTION(errors_to_use="pointwise")(
                self._data_chi2,
                self._model_chi2,
                self._pointwise_errors,
                self._par_vals,
                self._par_constraints,
                self._total_error_squared_log_sum,
            ),
        )

    def test_chi2_cov_mat(self):
        self.assertAlmostEqual(
            self._cost_chi2_cov_mat,
            self.CHI2_COST_FUNCTION(errors_to_use="covariance")(
                self._data_chi2,
                self._model_chi2,
                self._cov_mat_cholesky,
                None,
                None,
                self._cov_mat_log_det,
            ),
        )
        self.assertAlmostEqual(
            self._cost_chi2_cov_mat + self._par_cost,
            self.CHI2_COST_FUNCTION(errors_to_use="covariance")(
                self._data_chi2,
                self._model_chi2,
                self._cov_mat_cholesky,
                self._par_vals,
                self._par_constraints,
                self._cov_mat_log_det,
            ),
        )

    def test_nll_gaussian(self):
        self.assertAlmostEqual(
            self._cost_nll_gaussian,
            self.NLL_COST_FUNCTION(data_point_distribution="gaussian")(self._data_chi2, self._model_chi2, self._pointwise_errors, None, None),
        )
        self.assertAlmostEqual(
            self._cost_nll_gaussian + self._par_cost,
            self.NLL_COST_FUNCTION(data_point_distribution="gaussian")(
                self._data_chi2,
                self._model_chi2,
                self._pointwise_errors,
                self._par_vals,
                self._par_constraints,
            ),
        )

    def test_nll_poisson(self):
        self.assertAlmostEqual(
            self._cost_nll_poisson,
            self.NLL_COST_FUNCTION(data_point_distribution="poisson")(self._data_poisson, self._model_poisson, None, None),
        )
        self.assertAlmostEqual(
            self._cost_nll_poisson + self._par_cost,
            self.NLL_COST_FUNCTION(data_point_distribution="poisson")(self._data_poisson, self._model_poisson, self._par_vals, self._par_constraints),
        )

    def test_nllr_gaussian(self):
        self.assertAlmostEqual(
            self._cost_chi2_pointwise - self._total_error_squared_log_sum,
            self.NLL_COST_FUNCTION(data_point_distribution="gaussian", ratio=True)(
                self._data_chi2, self._model_chi2, self._pointwise_errors, None, None
            ),
        )
        self.assertAlmostEqual(
            self._cost_chi2_pointwise + self._par_cost - self._total_error_squared_log_sum,
            self.NLL_COST_FUNCTION(data_point_distribution="gaussian", ratio=True)(
                self._data_chi2,
                self._model_chi2,
                self._pointwise_errors,
                self._par_vals,
                self._par_constraints,
            ),
        )

    def test_nllr_poisson(self):
        self.assertAlmostEqual(
            self._cost_nllr_poisson,
            self.NLL_COST_FUNCTION(data_point_distribution="poisson", ratio=True)(self._data_poisson, self._model_poisson, None, None),
        )
        self.assertAlmostEqual(
            self._cost_nllr_poisson + self._par_cost,
            self.NLL_COST_FUNCTION(data_point_distribution="poisson", ratio=True)(
                self._data_poisson, self._model_poisson, self._par_vals, self._par_constraints
            ),
        )

    def test_gauss_approximation(self):
        self.assertAlmostEqual(
            self.NLL_COST_FUNCTION(data_point_distribution="poisson")(self._data_poisson_large, self._model_poisson_large, None, None),
            self.GAUSS_APPROXIMATION_COST_FUNCTION(errors_to_use="covariance")(
                self._data_poisson_large,
                self._model_poisson_large,
                np.zeros_like(self._cov_mat),
                None,
                None,
            )
            + 3 * np.log(2 * np.pi),
            places=3,
        )
        self.assertAlmostEqual(
            self.GAUSS_APPROXIMATION_COST_FUNCTION(errors_to_use="covariance")(
                self._data_poisson, self._model_poisson, np.zeros_like(self._cov_mat), None, None
            ),
            self.GAUSS_APPROXIMATION_COST_FUNCTION(errors_to_use="pointwise")(
                self._data_poisson,
                self._model_poisson,
                np.zeros_like(self._pointwise_errors),
                None,
                None,
            ),
        )
        self.assertAlmostEqual(
            self.CHI2_COST_FUNCTION(errors_to_use="covariance")(
                self._data_chi2,
                np.zeros_like(self._model_chi2),
                self._cov_mat_cholesky,
                None,
                None,
                self._cov_mat_log_det,
            ),
            self.GAUSS_APPROXIMATION_COST_FUNCTION(errors_to_use="covariance")(
                self._data_chi2, np.zeros_like(self._model_chi2), self._cov_mat, None, None
            ),
        )
        self.assertAlmostEqual(
            self.CHI2_COST_FUNCTION(errors_to_use="pointwise")(
                self._data_chi2,
                np.zeros_like(self._model_chi2),
                self._pointwise_errors,
                None,
                None,
                self._total_error_squared_log_sum,
            ),
            self.GAUSS_APPROXIMATION_COST_FUNCTION(errors_to_use="pointwise")(
                self._data_chi2, np.zeros_like(self._model_chi2), self._pointwise_errors, None, None
            ),
        )

    def test_chi2_raise(self):
        with self.assertRaises(ValueError):
            self.CHI2_COST_FUNCTION(errors_to_use="XYZ")
        with self.assertRaises(ValueError):
            self.CHI2_COST_FUNCTION(errors_to_use="covariance")(
                self._data_chi2,
                np.ones(10),
                self._cov_mat_cholesky,
                None,
                None,
                self._cov_mat_log_det,
            )
        with self.assertRaises(ValueError):
            self.CHI2_COST_FUNCTION(errors_to_use="covariance", fallback_on_singular=False)(
                self._data_chi2, self._model_chi2, None, None, None, self._cov_mat_log_det
            )
        with self.assertRaises(ValueError):
            self.CHI2_COST_FUNCTION(errors_to_use="pointwise", fallback_on_singular=False)(
                self._data_chi2,
                self._model_chi2,
                np.arange(len(self._pointwise_errors)),
                None,
                None,
                self._cov_mat_log_det,
            )

    def test_nll_raise(self):
        with self.assertRaises(ValueError):
            self.NLL_COST_FUNCTION(data_point_distribution="yes")

    def test_inf_cost(self):
        self.assertEqual(
            np.inf,
            self.CHI2_COST_FUNCTION(errors_to_use="covariance")(
                self._data_chi2,
                np.nan * np.ones_like(self._model_chi2),
                self._cov_mat_cholesky,
                None,
                None,
                self._cov_mat_log_det,
            ),
        )
        self.assertEqual(
            np.inf,
            self.CHI2_COST_FUNCTION(errors_to_use="pointwise")(
                self._data_chi2,
                np.nan * np.ones_like(self._model_chi2),
                self._pointwise_errors,
                None,
                None,
                self._cov_mat_log_det,
            ),
        )
        self.assertEqual(
            np.inf,
            self.CHI2_COST_FUNCTION(errors_to_use=None)(
                self._data_chi2,
                np.nan * np.ones_like(self._model_chi2),
                None,
                None,
                self._cov_mat_log_det,
            ),
        )
        self.assertEqual(
            np.inf,
            self.NLL_COST_FUNCTION("poisson", ratio=False)(self._data_poisson, -self._model_poisson, None, None),
        )
        self.assertEqual(
            np.inf,
            self.NLL_COST_FUNCTION("poisson", ratio=True)(self._data_poisson, -self._model_poisson, None, None),
        )
        self.assertEqual(
            np.inf,
            self.NLL_COST_FUNCTION("gaussian", ratio=False)(self._data_chi2, self._model_chi2, -self._pointwise_errors, None, None),
        )
        self.assertEqual(
            np.inf,
            self.NLL_COST_FUNCTION("gaussian", ratio=True)(self._data_chi2, self._model_chi2, -self._pointwise_errors, None, None),
        )


class TestCostBuiltinHist(TestCostBuiltin):
    CHI2_COST_FUNCTION = HistCostFunction_Chi2
    NLL_COST_FUNCTION = HistCostFunction_NegLogLikelihood


class TestCostBuiltinIndexed(TestCostBuiltin):
    CHI2_COST_FUNCTION = IndexedCostFunction_Chi2
    NLL_COST_FUNCTION = IndexedCostFunction_NegLogLikelihood


class TestCostBuiltinXY(TestCostBuiltin):
    CHI2_COST_FUNCTION = XYCostFunction_Chi2
    NLL_COST_FUNCTION = XYCostFunction_NegLogLikelihood


class TestCostUserDefined(unittest.TestCase):
    def setUp(self):
        def my_cost(a, b, c, d):
            return a**2 + (b + c) ** 2 + (d - 1) ** 2

        def my_cost_varargs(*args):
            return args[0] ** 2 + (args[1] + args[2]) ** 2 + (args[3] - 1) ** 2

        self._ref_par_vals = [1, 2, 5, 7]
        self._ref_cost = my_cost(*self._ref_par_vals)
        self._constraint = GaussianSimpleParameterConstraint(index=0, value=0, uncertainty=1)
        self._matrix_constraint = GaussianMatrixParameterConstraint(indices=[1], values=[0], matrix=[[1]])
        self._constraint_cost = self._ref_par_vals[0] ** 2 + self._ref_par_vals[1] ** 2
        self._ref_par_vals_constraints = self._ref_par_vals + [
            self._ref_par_vals,
            [self._constraint, self._matrix_constraint],
        ]

        self._cost_func = CostFunction(my_cost, arg_names=None, add_constraint_cost=False, add_determinant_cost=False)
        self._cost_func_constraints = CostFunction(my_cost, arg_names=None, add_constraint_cost=True, add_determinant_cost=False)
        self._cost_func_varargs = CostFunction(
            my_cost_varargs,
            arg_names=["a", "b", "c", "d"],
            add_constraint_cost=False,
            add_determinant_cost=False,
        )
        self._cost_func_varargs_constraints = CostFunction(
            my_cost_varargs,
            arg_names=["a", "b", "c", "d"],
            add_constraint_cost=True,
            add_determinant_cost=False,
        )

    def test_properties(self):
        self.assertEqual(self._cost_func.name, "my_cost")
        self.assertEqual(self._cost_func.arg_names, ["a", "b", "c", "d"])
        self.assertEqual(self._cost_func_varargs.name, "my_cost_varargs")
        self.assertEqual(self._cost_func.arg_names, ["a", "b", "c", "d"])

    def test_validate_raise(self):
        def _cost_args(*args):
            return np.sum(args)

        def _cost_kwargs(**kwargs):
            return np.sum(list(kwargs.values()))

        def _cost_bad_arg_name(cost):
            return cost

        _ = CostFunction(_cost_args, arg_names=["a", "b"])
        with self.assertRaises(ValueError):
            _ = CostFunction(_cost_args)
        with self.assertRaises(ValueError):
            _ = CostFunction(_cost_kwargs)
        with self.assertRaises(ValueError):
            _ = CostFunction(_cost_kwargs, arg_names=["a", "b"])
        with self.assertRaises(ValueError):
            _ = CostFunction(_cost_bad_arg_name)
        _ = CostFunction(_cost_bad_arg_name, arg_names=["a"])
        with self.assertRaises(ValueError):
            _ = CostFunction(_cost_args, arg_names=["cost"])

    def test_compare_cost(self):
        self.assertEqual(self._ref_cost, self._cost_func(*self._ref_par_vals))
        self.assertEqual(self._ref_cost, self._cost_func_constraints(*(self._ref_par_vals + [None, None])))
        self.assertEqual(
            self._ref_cost,
            self._cost_func_constraints(*(self._ref_par_vals + [self._ref_par_vals, None])),
        )
        self.assertEqual(
            self._ref_cost + self._constraint_cost,
            self._cost_func_constraints(*self._ref_par_vals_constraints),
        )

    def test_compare_cost_varargs(self):
        self.assertEqual(self._ref_cost, self._cost_func_varargs(*self._ref_par_vals))
        self.assertEqual(
            self._ref_cost,
            self._cost_func_varargs_constraints(*(self._ref_par_vals + [None, None])),
        )
        self.assertEqual(
            self._ref_cost,
            self._cost_func_varargs_constraints(*(self._ref_par_vals + [self._ref_par_vals, None])),
        )
        self.assertEqual(
            self._ref_cost + self._constraint_cost,
            self._cost_func_varargs_constraints(*self._ref_par_vals_constraints),
        )



# flake8: noqa F401 (imported but unused)
from collections import OrderedDict
from copy import deepcopy

import numpy as np

from ...tools import print_dict_as_table
from .._base import DataContainerBase, FitBase
from .._base.cost import CostFunction_Chi2
from ..util import collect
from .container import IndexedContainer
from .model import IndexedModelFunction, IndexedParametricModel
from .plot import IndexedPlotAdapter

__all__ = ["IndexedFit"]


class IndexedFit(FitBase):
    CONTAINER_TYPE = IndexedContainer
    MODEL_TYPE = IndexedParametricModel
    MODEL_FUNCTION_TYPE = IndexedModelFunction
    PLOT_ADAPTER_TYPE = IndexedPlotAdapter
    RESERVED_NODE_NAMES = {
        "data",
        "model",
        "cost",
        "data_error",
        "model_error",
        "total_error",
        "data_cov_mat",
        "model_cov_mat",
        "total_cov_mat",
        "data_cor_mat",
        "model_cor_mat",
        "total_cor_mat",
    }
    _BASIC_ERROR_NAMES = {"data_error", "model_error", "data_cov_mat", "model_cov_mat"}

    def __init__(
        self,
        data,
        model_function,
        cost_function="chi2",
        minimizer=None,
        minimizer_kwargs=None,
        dynamic_error_algorithm="nonlinear",
    ):
        """
        Construct a fit of a model to a series of indexed measurements.

        :param data: the measurement values
        :type data: iterable of float
        :param model_function: the model function
        :type model_function: :py:class:`~kafe2.fit.indexed.IndexedModelFunction` or unwrapped native Python function
        :param cost_function: the cost function
        :type cost_function: :py:class:`~kafe2.fit._base.CostFunctionBase`-derived or unwrapped native Python function
        :param minimizer: the minimizer to use for fitting.
        :type minimizer: None, "iminuit", "tminuit", or "scipy".
        :param minimizer_kwargs: dictionary with kwargs for the minimizer.
        :type minimizer_kwargs: dict
        """
        super(IndexedFit, self).__init__(
            data=data,
            model_function=model_function,
            cost_function=cost_function,
            minimizer=minimizer,
            minimizer_kwargs=minimizer_kwargs,
            dynamic_error_algorithm=dynamic_error_algorithm,
        )

    # -- private methods

    def _init_nexus(self):
        super(IndexedFit, self)._init_nexus()

        self._nexus.add_alias("y_data", alias_for="data")
        self._nexus.add_alias("y_model", alias_for="model")
        self._nexus.add_alias("y_total_cov_mat", alias_for="total_cov_mat")

        self._nexus.add_dependency("model", depends_on=("parameter_values"))

    def _report_data(self, output_stream, indent, indentation_level):
        output_stream.write(indent * indentation_level + "########\n")
        output_stream.write(indent * indentation_level + "# Data #\n")
        output_stream.write(indent * indentation_level + "########\n\n")

        _data_table_dict = OrderedDict()
        _data_table_dict["Index"] = range(self.data_size)
        _data_table_dict["Data"] = self.data
        if self.has_data_errors:
            _data_table_dict["Data Error"] = self.data_error
            _data_table_dict["Data Correlation Matrix"] = self.data_cor_mat

        print_dict_as_table(_data_table_dict, output_stream=output_stream, indent_level=indentation_level + 1)
        output_stream.write("\n")

    def _report_model(self, output_stream, indent, indentation_level):
        # call base method to show header and model function
        super(IndexedFit, self)._report_model(output_stream, indent, indentation_level)
        _model_table_dict = OrderedDict()
        _model_table_dict["Index"] = range(self.data_size)
        _model_table_dict["Model"] = self.model
        if self.has_model_errors:
            _model_table_dict["Model Error"] = self.model_error
            _model_table_dict["Model Correlation Matrix"] = self.model_cor_mat

        print_dict_as_table(_model_table_dict, output_stream=output_stream, indent_level=indentation_level + 1)
        output_stream.write("\n")

        if self._param_model.get_matching_errors({"relative": True}):
            output_stream.write(indent * (indentation_level + 1))
            output_stream.write("Model covariance matrix was calculated dynamically relative to y model values.\n")
            output_stream.write("\n")

    # -- public properties

    def _set_new_data(self, new_data):
        if isinstance(new_data, self.CONTAINER_TYPE):
            self._data_container = deepcopy(new_data)
        elif isinstance(new_data, DataContainerBase):
            raise TypeError("Incompatible container type '%s' (expected '%s')" % (type(new_data), self.CONTAINER_TYPE))
        else:
            self._data_container = IndexedContainer(new_data, dtype=float)
        self._data_container._on_error_change_callback = self._on_error_change

        self._nexus.get("data").mark_for_update()

    def _set_new_parametric_model(self):
        self._param_model = IndexedParametricModel(self._model_function, self.parameter_values, shape_like=self.data)

    def _get_default_x_ticks(self):
        if self.data_size <= 10:
            return np.arange(self.data_size)
        else:
            return []

    # -- public properties

    @property
    def model(self):
        """array of model predictions for the data points"""
        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
        return self._param_model.data



import numpy as np

# TODO documentation

__all__ = [
    "linear_model",
    "linear_model_derivative",
    "quadratic_model",
    "quadratic_model_derivative",
    "cubic_model",
    "cubic_model_derivative",
    "exponential_model",
    "normal_distribution",
]


def linear_model(x, a, b):
    return a * x + b


def linear_model_derivative(x, a, b):
    return np.ones_like(x) * a


linear_model.expression_format_string = "{a} * {x} + {b}"
linear_model.latex_expression_format_string = r"{a} \, {x} + {b}"


def quadratic_model(x, a, b, c):
    return a * x**2 + b * x + c


def quadratic_model_derivative(x, a, b, c):
    return 2 * a * x + b


quadratic_model.expression_format_string = "{a} * {x} ** 2 + {b} * {x} + {c}"
quadratic_model.latex_expression_format_string = r"{a} \, {x}^2 + {b} \, {x} + {c}"


def cubic_model(x, a, b, c, d):
    return a * x**3 + b * x**2 + c * x + d


def cubic_model_derivative(x, a, b, c, d):
    return 3 * a * x**2 + 2 * b * x + c


cubic_model.expression_format_string = "{a} * {x} ** 3 + {b} * {x} ** 2 + {c} * {x} + {d}"
cubic_model.latex_expression_format_string = r"{a} \, {x}^3 + {b} \, {x}^2 + {c} \, {x} + {d}"


def exponential_model(x, A_0, x_0):
    return A_0 * np.exp(x / x_0)


def exponential_model_derivative(x, A_0, x_0):
    return A_0 * np.exp(x / x_0)


exponential_model.expression_format_string = "{A_0} * exp({x} / {x_0})"
exponential_model.latex_expression_format_string = r"{A_0} \, e^{{{x} \ / {x_0}}}"


def normal_distribution(x, mu, sigma):
    return np.exp(-0.5 * ((x - mu) / sigma) ** 2) / np.sqrt(2.0 * np.pi * sigma**2)


normal_distribution.expression_format_string = "1 / sqrt(2 * pi * {sigma}) * exp(-0.5 * (({x} - {mu}) / {sigma}) ** 2)"
normal_distribution.latex_expression_format_string = (
    r"\frac{{1}}{{2 \pi {sigma}}} e^{{- \frac{{1}}{{2}}" r"\left( \frac{{{x} - {mu}}}{{{sigma}}} \right)^2 }}"
)

normal_distribution_pdf = normal_distribution  # Backwards compatibility


STRING_TO_FUNCTION = {
    "line": linear_model,
    "linear": linear_model,
    "linear_model": linear_model,
    "quadratic": quadratic_model,
    "quadratic_model": quadratic_model,
    "cubic": cubic_model,
    "cubic_model": cubic_model,
    "exp": exponential_model,
    "exponential": exponential_model,
    "exponential_model": exponential_model,
    "normal": normal_distribution,
    "normal_distribution": normal_distribution,
    "normal_distribution_pdf": normal_distribution,
}
